/*
	************************************************************
	2D RLE encoded scanlines demo for iLCD panels
	Copyright (c) 2013 by demmel products gmbh  - www.demmel.com
	************************************************************
																
	This is a simple test code to show how to work with 
	run-lenght encoded read / write scanlines on iLCD devices.

	It reads an area defined via AREA_START_X, AREA_START_Y,
	AREA_WIDTH and AREA_HEIGHT	via the appropriate commands,
	inverts the array of pixel data and writes back the data
	to the same location. The commands used for this are
	\iDNr and \iDNw. Have a look at iLCD Manager XE
	to see which paramters are required by these commands.
	
	The "real stuff" is done in file 2D_RLE.c, this file calls 
	the appropriate functions only to show how to work with 
	1D / 2D encoded scanlines.

	Please note that this code will work with DPC3050 and greater
	iLCD controllers only, the firmware version must be v4.1 or greater.

	Communication with the iLCD panel is made via serial port only, 
	port settings are hard-coded here in OpenPort().

	If you want to learn more about how to generally talk to an iLCD
	device (e.g. vai USB or Ethernet) check out the iLoader project instead.
*/

//---------------------------------------------------------------------------
#include <stdio.h>
#pragma hdrstop

#include <tchar.h>
#include "2D_RLE.h"
//---------------------------------------------------------------------------
#pragma argsused

#define USE_SERIAL_PORT						1		// use serial port communcation - see at OpenPort()
#define PROCESS_SINGLE_INCOMING_BYTES  1		// if true, no buffer is needed 
#define INTERLACED_VALUE					1

#if USE_SERIAL_PORT
	#include "serial.h"
#endif	

#define CID			0xAA
#define ACK			0x06

#define AREA_START_X	20
#define AREA_START_Y	10
#define AREA_WIDTH	200
#define AREA_HEIGHT	(100 / INTERLACED_VALUE * INTERLACED_VALUE)	// to keep reading simple

static byte compressed_data[AREA_WIDTH * 2 + 2];

#define ERR_MSG_OUT(s)	printf(s)

//---------------------------------------------------------------------------
bool OpenPort(void)
{
#if USE_SERIAL_PORT
	return Serial_Open("com1:", 115200, false);
#else	
	// *** dummy code ***

	// Please fill in your code here to setup the selected port (e.g. serial port)
	// the iLCD device is connected to.

	return false;
#endif
}  
//---------------------------------------------------------------------------
bool ClosePort(void)
{
#if USE_SERIAL_PORT
	return Serial_Close();
#else	
	// *** dummy code ***

	// Please fill in your code here to setup the selected port (e.g. serial port)
	// the iLCD device is connected to.

	return false;
#endif
}  
//---------------------------------------------------------------------------
bool GetByteFromDevice(byte *b)
{
#if USE_SERIAL_PORT
	return Serial_ReceiveChar(b);
#else	
	// *** dummy code ***
	
	// Please fill in your code here to get a byte from the selected port (e.g. serial port)
	// from the iLCD device and to write it into 'b'

	return false;
#endif
}      
//---------------------------------------------------------------------------
bool SendByteToDevice(byte b)
{
#if USE_SERIAL_PORT
	return Serial_Transmit(&b, 1);
#else	
	// *** dummy code ***
	
	// Please fill in your code here to send a byte to the selected port (e.g. serial port)
	// the iLCD device is connected to.
	
	return false;
#endif
}    
//---------------------------------------------------------------------------
bool SendDataToDevice(byte *data, int length)
{
	int i;
	byte b;

	for (i = 0; i < length; i++)
	{
		b = *data++;
		
		if (b == CID)
		{
			// quote CIDs
			if (!SendByteToDevice(b))
				return false;
		}      

		if (!SendByteToDevice(b))
			return false;
	}                                

	return true;
}  
//---------------------------------------------------------------------------
bool SendCommandToDevice(byte *data, int length)
{
	byte b;
	
	if (!SendByteToDevice(CID))
		return false;
	
	if (!SendDataToDevice(data, length))
		return false;

	if (!GetByteFromDevice(&b))
		return false;
		
	if (b == ACK)
	{
		// got acknowledge
		return true;
	}

	return false;
}
//---------------------------------------------------------------------------
bool GotoXY(int x, int y)
{
	byte cmd[20] = "CK";
	int i = strlen(cmd);

	cmd[i++] = (x >> 8) & 0xFF;
	cmd[i++] = x & 0xFF;

	cmd[i++] = (y >> 8) & 0xFF;
	cmd[i++] = y & 0xFF;

	return SendCommandToDevice(cmd, i);
}
//---------------------------------------------------------------------------
// previous_scanline is the scanline previously (at y + scanline_offset) got - if NULL, 1D is used
// scanline must be able to hold at least 'width' * 2 bytes                                
bool Read_RLE_Scanline(int x, int y, int scanline_offset, word *scanline, word *previous_scanline, int width)
{
#if PROCESS_SINGLE_INCOMING_BYTES
	int pcnt, pix_out_count = 0;
#endif	

	byte cmd[20] = "DN";
	int i = strlen(cmd);
	byte b;
	int cnt;

	if (!GotoXY(x, y))
		return false;

	cmd[i++] = 'r';
	cmd[i++] = (signed char) scanline_offset;

	cmd[i++] = (width >> 8) & 0xFF;
	cmd[i++] = width & 0xFF;

	if (!SendCommandToDevice(cmd, i))
		return false;

	if (!GetByteFromDevice(&b))
		return false;
		
	cnt = ((word) b) << 8;

	if (!GetByteFromDevice(&b))
		return false;
	
	cnt |= b;

	// Under worst case situations the compressed data size is 2 byte more than
	// the number of pixels * 2 to be read.
	
	if (cnt > (width * 2) + 2)
		return false;

	// now we have the # of bytes to be read from the device

#if PROCESS_SINGLE_INCOMING_BYTES
	InitDecode(scanline, previous_scanline, width);	
#endif	

	for (i = 0; i < cnt; i++)
	{
		if (!GetByteFromDevice(&b))
			return false;

#if PROCESS_SINGLE_INCOMING_BYTES
		pcnt = DecodeByte(b);

		if (pcnt < 0)
		{
			// error in data
			return false;
		}                  
		
		pix_out_count += pcnt;

		if (pix_out_count > width)
		{
			// this should *never* happen !!
			return false;
		}
#else
		compressed_data[i] = b;
#endif
	}

	// now get the terminating ACK
	if (!GetByteFromDevice(&b))
		return false;

	if (b != ACK)
		return false;
			 
#if PROCESS_SINGLE_INCOMING_BYTES
	return pix_out_count == width;
#else
	return RunLengthDecode(compressed_data, cnt, scanline, previous_scanline, width);		
#endif
}            
//---------------------------------------------------------------------------
// Read screen area starting from x / y
// 'word *data' must be able to hold at least 'width' * 'height' words that
// is width * height * 2 bytes
bool ReadScreenArea(int x, int y_start, word *data, int width, int height)
{
	int y, i;
	int inc = INTERLACED_VALUE;
	word *act_scanline;
	word *prev_scanline = NULL;

	// to keep the sample code simple we assume ((height % inc) == 0)
	
	for (i = 0; i < inc; i++)
	{
		for (y = y_start + i; y < y_start + height; y += inc)
		{
			act_scanline = data + ((y - y_start) * width);

			if (y >= y_start + inc)
				prev_scanline = data + ((y - y_start - inc) * width);

			if (!Read_RLE_Scanline(x, y, (y >= y_start + inc) ? -inc : 0, act_scanline, prev_scanline, width))
				return false;
		}
	}

	return true;
}
//---------------------------------------------------------------------------
// previous_scanline is the scanline previously (at y + scanline_offset) got - if NULL, 1D is used
bool Write_RLE_Scanline(int x, int y, int scanline_offset, word *scanline, word *previous_scanline, int width)
{
	byte cmd[20] = "DNw";
	int i = strlen(cmd);
	byte b;
	int cnt;

	if (!GotoXY(x, y))
		return false;

	/*	
		Note: On systems with very low memory space using an extra array for 
		'compressed_data' can be a problem (it must be able to hold at least 
		max. screen width * 2 + 2 bytes == max. 2050 bytes for a 1024 pixel display). 
	  
		To bypass this memory requirement a really dirty trick can be used 
		for single scanlines (not applicable on this sample code):
		
		The scanline data array and the compressed_data array can share
		the same memory (!), if the pointer to compressed_data is set
		to a minimum of 3 bytes before the start of scanline data.

		So the commonly shared array needs to have a size of only 
		max. screen width * 2 + 3 bytes in this case, RunLengthEncode() internal's 
		write pointer will *never* overtake the read pointer.

		Example:
		
		byte shared_data[AREA_WIDTH * 2 + 3];
		
		compressed_data = shared_data;
		scanline = shared_data + 3;

		Calling routine must fill in the scanline data at shared_data + 3 before
		calling this routine.
	*/
	
	cnt = RunLengthEncode(compressed_data, scanline, previous_scanline, width);

	cmd[i++] = (signed char) scanline_offset;
	cmd[i++] = (width >> 8) & 0xFF;
	cmd[i++] = width & 0xFF;
		
	cmd[i++] = (cnt >> 8) & 0xFF;
	cmd[i++] = cnt & 0xFF;

	if (!SendByteToDevice(CID))
		return false;
	
	if (!SendDataToDevice(cmd, i))
		return false;

	if (!SendDataToDevice(compressed_data, cnt))
		return false;

	// get the ACK returned from the device
	if (!GetByteFromDevice(&b))
		return false;

	if (b != ACK)
		return false;
	       
	return true;
}            
//---------------------------------------------------------------------------
// Write screen area with size 'width' and 'height' to x / y
bool WriteScreenArea(int x, int y, word *data, int width, int height)
{
	int i;
	word *act_scanline;
	word *prev_scanline = NULL;

	for (i = y; i < y + height; i++)
	{
		act_scanline = data + ((i - y) * width);

		if (i > y)
			prev_scanline = data + ((i - y - 1) * width);

		if (!Write_RLE_Scanline(x, i, (i > y) ? -1 : 0, act_scanline, prev_scanline, width))
			return false;
	}

	return true;
}
//---------------------------------------------------------------------------
int _tmain(int argc, _TCHAR* argv[])
{
	int i;
	word area[AREA_WIDTH * AREA_HEIGHT];
	
	if (!OpenPort())
	{
		ERR_MSG_OUT("Initialize failed");
		return 1;
	}

	if (!ReadScreenArea(AREA_START_X, AREA_START_Y, area, AREA_WIDTH, AREA_HEIGHT))
	{
		ClosePort();
		ERR_MSG_OUT("ReadScreenArea() failed");
		return 1;                        
	}

	// invert area just read
	for (i = 0; i < AREA_WIDTH * AREA_HEIGHT; i++)
		area[i] = ~(area[i]);

	if (!WriteScreenArea(AREA_START_X, AREA_START_Y, area, AREA_WIDTH, AREA_HEIGHT))
	{
		ClosePort();
		ERR_MSG_OUT("WriteScreenArea() failed");
		return 1;                        
	}
																											  
	ClosePort();
	return 0;
}
//---------------------------------------------------------------------------
